import { assertNever } from "@fern-api/core-utils";
import { Reference } from "./Reference";
import { Type } from "./Type";
import { AstNode } from "./core/AstNode";
export class TypeInstantiation extends AstNode {
    internalType;
    constructor(internalType) {
        super();
        this.internalType = internalType;
    }
    static int(value) {
        return new this({ type: "int", value });
    }
    static float(value) {
        return new this({ type: "float", value });
    }
    static bool(value) {
        return new this({ type: "bool", value });
    }
    static str(value, config = {
        multiline: false,
        startOnNewLine: false,
        endWithNewLine: false
    }) {
        return new this({ type: "str", value, config });
    }
    static bytes(value) {
        return new this({ type: "bytes", value });
    }
    static list(values, config = { endWithComma: false }) {
        const list = new this({ type: "list", values, config });
        values.forEach((value) => list.inheritReferences(value));
        return list;
    }
    static set(values, config = { endWithComma: false }) {
        const set = new this({ type: "set", values, config });
        values.forEach((value) => set.inheritReferences(value));
        return set;
    }
    static tuple(values, config = { endWithComma: false }) {
        const tuple = new this({ type: "tuple", values, config });
        values.forEach((value) => tuple.inheritReferences(value));
        return tuple;
    }
    static dict(entries, config = { endWithComma: false }) {
        const dict = new this({ type: "dict", entries, config });
        entries.forEach((entry) => {
            dict.inheritReferences(entry.key);
            dict.inheritReferences(entry.value);
        });
        return dict;
    }
    static none() {
        return new this({ type: "none" });
    }
    static uuid(value) {
        const uuid = new this({ type: "uuid", value });
        uuid.addReference(new Reference({ name: "UUID", modulePath: ["uuid"] }));
        return uuid;
    }
    write(writer) {
        switch (this.internalType.type) {
            case "int":
                writer.write(this.internalType.value.toString());
                break;
            case "float":
                writer.write(this.internalType.value.toString());
                break;
            case "bool":
                if (this.internalType.value) {
                    writer.write("True");
                }
                else {
                    writer.write("False");
                }
                break;
            case "str":
                if (this.internalType.config?.multiline) {
                    const { startOnNewLine, endWithNewLine } = this.internalType.config;
                    this.writeStringWithTripleQuotes({
                        writer,
                        value: this.internalType.value,
                        startOnNewLine,
                        endWithNewLine
                    });
                }
                else {
                    writer.write(`"${this.escapeString(this.internalType.value)}"`);
                }
                break;
            case "bytes":
                writer.write(`b"${this.internalType.value}"`);
                break;
            case "list": {
                const internalType = this.internalType;
                writer.write("[");
                internalType.values.forEach((value, index) => {
                    if (index > 0) {
                        writer.write(", ");
                    }
                    value.write(writer);
                    if (index === internalType.values.length - 1 && internalType.config?.endWithComma) {
                        writer.write(",");
                    }
                });
                writer.write("]");
                break;
            }
            case "set": {
                const internalType = this.internalType;
                writer.write("{");
                internalType.values.forEach((value, index) => {
                    if (index > 0) {
                        writer.write(", ");
                    }
                    value.write(writer);
                    if (index === internalType.values.length - 1 && internalType.config?.endWithComma) {
                        writer.write(",");
                    }
                });
                writer.write("}");
                break;
            }
            case "tuple": {
                const internalType = this.internalType;
                writer.write("(");
                this.internalType.values.forEach((value, index) => {
                    if (index > 0) {
                        writer.write(", ");
                    }
                    value.write(writer);
                    if (
                    // If the tuple is of length 1, then we must always add a trailing comma
                    internalType.values.length === 1 ||
                        // Otherwise, check the config that was specified
                        (index === internalType.values.length - 1 && internalType.config?.endWithComma)) {
                        writer.write(",");
                    }
                });
                writer.write(")");
                break;
            }
            case "dict": {
                const internalType = this.internalType;
                writer.write("{");
                internalType.entries.forEach((entry, index) => {
                    if (index > 0) {
                        writer.write(", ");
                    }
                    entry.key.write(writer);
                    writer.write(": ");
                    entry.value.write(writer);
                    if (index === internalType.entries.length - 1 && internalType.config?.endWithComma) {
                        writer.write(",");
                    }
                });
                writer.write("}");
                break;
            }
            case "none":
                writer.write("None");
                break;
            case "uuid":
                writer.write(`UUID("${this.internalType.value}")`);
                break;
            default:
                assertNever(this.internalType);
        }
    }
    writeStringWithTripleQuotes({ writer, value, startOnNewLine, endWithNewLine }) {
        writer.write('"""');
        const lines = value.split("\n");
        // If there is only one line, we can just write it as a single line string
        if (lines.length <= 1) {
            writer.write(this.escapeString(lines[0] ?? ""));
            writer.write('"""');
            return;
        }
        if (startOnNewLine) {
            writer.writeNoIndent("\\\n");
        }
        lines.forEach((line, idx) => {
            writer.writeNoIndent(this.escapeString(line));
            // If this is the last line, add a newline escape
            if (idx === lines.length - 1) {
                if (endWithNewLine) {
                    writer.writeNoIndent("\\\n");
                }
            }
            else {
                writer.writeNoIndent("\n");
            }
        });
        writer.writeNoIndent('"""');
    }
    /**
     * Escapes certain special characters if they're NOT already preceded
     * by a backslash. Specifically:
     *
     *   - "  -> \"
     *   - '  -> \'
     *   - \  -> \\
     *   - literal \t -> \t
     *   - literal \n -> \n
     *   - literal \r -> \r
     *
     * Uses a negative lookbehind `(?<!\\)` to handle consecutive matches like
     * \n\n correctly, since each \n is independently matched in the original string.
     *
     * @param input The input string to be escaped
     */
    escapeString(input) {
        // Negative lookbehind ensures the character is NOT preceded by a backslash
        // in the original string.
        const pattern = /(?<!\\)(["'\\\t\n\r])/g;
        const replacements = {
            '"': '\\"',
            "'": "\\'",
            "\\": "\\\\",
            "\t": "\\t",
            "\n": "\\n",
            "\r": "\\r"
        };
        return input.replace(pattern, (char) => replacements[char] ?? char);
    }
}
export { Type };
